#!/usr/bin/python
# -*- coding: utf-8 -*-
#Copyright (C) 2009 Fabricio Silva

"""
Module de POD
liée au stage de Benjamin Vial 
(encadré par Sergio Bellizzi et Christophe Vergez)
"""

import warnings
import numpy as np
from Solveur_dim import SimulationTemporelle, systeme, acous, time
from scipy import integrate

debug = False

def systeme_py_pom(inc,t,sn,cn,pmpts,ypts,matp,mx):
    x = np.dot(matp, inc)
    dx = systeme.systeme(x+mx,t,sn,cn,pmpts,ypts)
    return np.dot(matp.T, dx)

def jacobien_py_pom(inc,t,sn,cn,pmpts,ypts,matp,mx):
    x = np.dot(matp, inc)
    Jf = systeme.jacobien(x+mx,t,sn,cn,pmpts,ypts)
    return np.dot(matp.T, np.dot(Jf,matp))

    
def calcule_pom(Xsol, sort=True):
    mX = np.mean(Xsol, axis=0)
    sigma = np.cov((Xsol-mX).T, bias=1)
    lambdan, psin = np.linalg.eig(sigma)
    if sort:
        idx = lambdan.argsort()[::-1]
        return lambdan[idx], psin[:,idx], mX[idx]
    else:
        return lambdan, psin, mX

def trace_repartition_pom(matP, Fig=None):
    if Fig==None:
        import matplotlib.pyplot as plt
        Fig = plt.figure()
    N1, N2 = int(matP.shape[0]/2-1), matP.shape[1]
    width = 0.7
    lLabels = np.array([['$Re p_{%d}$'%n, '$Im p_{%d}$'%n] for n in xrange(N1)]).ravel().tolist()
    lLabels.extend(['$h$', '$dh/dt$'])
    for n in xrange(N2):
        S = Fig.add_subplot(N2,1,n+1)
        S.bar(np.arange(2*N1+2), np.abs(matP[:,n]), 0.7)
        S.set_ylabel(r'$Y_{%d}$'%n)
        S.set_ylim((0,1))
        S.set_yticks((0,1))
        S.set_xticks(np.arange(2*N1+2)+width/2.)
        S.set_xticklabels(lLabels)
        S.label_outer()
    return Fig

class SimuPOM(SimulationTemporelle):
    def __init__(self, **kwargs):
        self.matP = kwargs.pop('matP', None)
        self.mX = kwargs.pop('mX', None)
        SimulationTemporelle.__init__(self, **kwargs)
        self.nb_pom = self.matP.shape[1]
        if self.matP.shape[0]!= 2+2*self.sn.shape[0]:
            print self.matP.shape[0], 2*self.sn.shape[0]+2
            raise ValueError, "Dimensions de la matrice P non conformes à l'impédance."
    
    def conditions_initiales(self, **kwargs):
        Y0 = kwargs.pop('X0', None)
        if Y0==None:
            SimulationTemporelle.conditions_initiales(self, **kwargs)
            self.inc0 = np.dot(self.matP.T, self.inc0-self.mX)
        else:
            self.inc0 = np.atleast_1d(np.dot(self.matP.T, Y0-self.mX))
        
    def solveur_pom(self, raffine=1., implementation='NumPy'):
        u"""
        Résoud le système différentiel associé aux POM.
        On fait ici appels au système (et jacobien) basique
        après transformation par la matrice de passage.
        """
        systeme.param.ktable = self.ktable
        systeme.param.qtable = self.qtable
        systeme.param.k = self.k
        systeme.param.wr = self.wr
        systeme.param.qr = self.qr
        systeme.param.rho = acous.rho
        systeme.param.w = self.W
        temps = np.linspace(self.vectemps[0], self.vectemps[-1], \
            raffine*len(self.vectemps))
        self.grandeurs_controle()
        
        if implementation.lower()=="numpy":
            fct_sys = systeme_py_pom
            fct_jac = jacobien_py_pom
        elif implementation.lower()=="fortran":
            fct_sys = systeme.systeme_pom
            fct_jac = systeme.jacobien_pom
        else:
            raise IOError, u"Quelle implémentation utiliser pour la simu POM?"
            
        if debug: print u"Calcul des auto-oscillations pomales"
        # REMARQUE : 
        # faire les opérations matricielles en fortran est plus long 
        # qu'utiliser les routines numpy/blas optimisées.
        Sol, dic = integrate.odeint(fct_sys, \
            self.inc0, temps, \
            args=(self.sn, self.Cn, \
                self.profil_pbouche.pts, \
                self.profil_ouverture.pts, self.matP, self.mX), \
            full_output=1, h0=1./self.Fe, hmax=1.2/self.Fe, \
            Dfun=fct_jac, \
            printmessg=True, mxstep=5000, ixpr=False)
        self.YSol = Sol[::raffine,:]
        self.Sol = np.dot(self.YSol, self.matP.T)+self.mX
        self.label = time.strftime("SimuPOM.%Y%m%d.%Hh%M")
        return dic
